import webbrowser
from pathlib import Path
from typing import Callable, List, cast

from simple_term_menu import TerminalMenu

IS_CURSOR_CYCLING = True
IS_SCREEN_CLEARED = True


def _directory_list_menu(directory_list: List[str]) -> int:
    final_dir_list = directory_list
    if "[q] Quit" not in final_dir_list:
        final_dir_list.append("[q] Quit")
    directory_list_menu = TerminalMenu(
        menu_entries=final_dir_list,
        title="Choose a directory",
        cycle_cursor=IS_CURSOR_CYCLING,
        clear_screen=IS_SCREEN_CLEARED,
        # preview_command="",
        # preview_size=0,
        show_search_hint=True,
        show_shortcut_hints=True,
    )
    selected_directory = directory_list_menu.show()
    return cast(int, selected_directory)


def _file_list_menu(directory: str, file_list: List[str]) -> int:
    if "[;] Back" not in file_list:
        file_list.append("[;] Back")
        file_list.append("[q] Quit")
    file_list_menu = TerminalMenu(
        menu_entries=file_list,
        title=f"Choose a file from {directory}",
        cycle_cursor=IS_CURSOR_CYCLING,
        clear_screen=IS_SCREEN_CLEARED,
        # preview_command="",
        # preview_size=0,
        show_search_hint=True,
        show_shortcut_hints=True,
    )
    selected_file = file_list_menu.show()
    return cast(int, selected_file)


def _confirmation_menu(choosen_file: str, directory: str) -> int:
    confimation_menu = TerminalMenu(
        title=f"Are you sure you want to choose {directory}/{choosen_file}?"
        f" (This will open a web browser tab to open a new issue)",
        menu_entries=["YES", "NO", "[;] Back", "[q] Quit"],
        cycle_cursor=IS_CURSOR_CYCLING,
        clear_screen=IS_SCREEN_CLEARED,
        show_search_hint=True,
        show_shortcut_hints=True,
    )
    choice = confimation_menu.show()
    return cast(int, choice)


def get_dir_list(repo_path: Path, ignore_matches: Callable[[str], bool]) -> List[str]:
    return list(
        set(
            [
                file.parent.name
                for file in repo_path.rglob("*.po")
                if not ignore_matches(str(file))
            ]
        )
    )


def get_files_from_dir(
    directory: str, repo_path: Path, ignore_matches: Callable[[str], bool]
) -> List[str]:
    path = Path(str(repo_path) + "/" + directory)
    return [file.name for file in path.rglob("*.po") if not ignore_matches(str(file))]


def interactive_output(path: Path, ignore_matches: Callable[[str], bool]) -> None:
    directory_options = get_dir_list(repo_path=path, ignore_matches=ignore_matches)
    while True:
        selected_dir = _directory_list_menu(directory_options)
        if selected_dir == (len(directory_options) - 1):
            exit(0)
        directory = directory_options[selected_dir]
        file_options = get_files_from_dir(
            directory=directory, repo_path=path, ignore_matches=ignore_matches
        )
        # TODO: Add stats on files and also add reservations
        selected_file = _file_list_menu(directory, file_options)
        if selected_file == (len(file_options) + 1):
            exit(0)
        elif selected_file == len(file_options):
            continue
        file = file_options[selected_file]
        final_choice = _confirmation_menu(file, directory)
        if final_choice == 3:
            exit(0)
        elif final_choice == 2:
            continue
        else:
            break
    if final_choice == 0:
        webbrowser.open(
            f"https://git.afpy.org/AFPy/python-docs-fr/issues/new?title=Je%20travaille%20sur%20"
            f"{directory}/{file}"
            f"&body=%0A%0A%0A---%0AThis+issue+was+created+using+potodo+interactive+mode."
        )
    else:
        exit()



from contextlib import suppress
from pathlib import Path

import pytest

from potodo.potodo import main


@pytest.fixture(name="repo_dir")
def _repo_dir():
    return Path(__file__).resolve().parent / "fixtures" / "repository"


@pytest.fixture
def run_potodo(repo_dir, capsys, monkeypatch):
    def run_it(argv):
        monkeypatch.setattr(
            "sys.argv", ["potodo", "--no-cache", "-p", str(repo_dir)] + argv
        )
        with suppress(SystemExit):
            main()
        return capsys.readouterr()

    return run_it



import logging
import re
from datetime import datetime
from typing import Any, Dict, List, Tuple
from urllib.parse import urlparse

import requests


def get_issue_reservations(api_url: str) -> Dict[str, Tuple[Any, Any]]:
    """Will get the repository name then request all the issues and put them in a dict"""

    logging.info("Getting issue reservations from %s", urlparse(api_url).hostname)
    issues: List[Dict[Any, Any]] = []
    logging.debug("Getting %s", api_url)
    next_url = api_url
    while next_url:
        logging.debug("Getting %s", next_url)
        resp = requests.get(next_url)
        if resp.status_code == 403:
            # Rate limit exceeded
            return {}
        issues.extend(resp.json())
        next_url = resp.links.get("next", {}).get("url")

    resp = requests.get(api_url)
    if resp.status_code == 403:
        # Rate limit exceeded
        return {}
    issues.extend(resp.json())
    logging.debug("Found %s issues", len(issues))

    reservations = {}

    for issue in issues:
        # Maybe find a better way for not using python 3.8 ?
        yes = re.search(r"\w*/[\w\-\.]*\.po", issue["title"])
        if yes:
            creation_date = datetime.strptime(
                issue["created_at"].split("T")[0], "%Y-%m-%d"
            ).date()
            reservations[yes.group()] = (issue["user"]["login"], creation_date)

    logging.debug("Found %s reservations", len(reservations))
    return reservations



from pathlib import Path

from potodo.potodo import main

REPO_DIR = Path(__file__).resolve().parent / "fixtures" / "repository"
GIT_REPO_DIR = Path(__file__).resolve().parent / "fixtures" / "git_repository"


def test_git(capsys, monkeypatch):
    """Ensure than excluded files are **not** parsed.

    Parsing excluded files can lead to surprises, here, parsing a
    `.po` file in `.git` may not work, it may just be a branch or
    whatever and contain a sha1 instead.

    I name it dotgit instead of .git, to not scare git.
    """
    monkeypatch.setattr(
        "sys.argv", ["potodo", "-p", str(GIT_REPO_DIR), "--exclude", "dotgit/"]
    )
    main()
    out, err = capsys.readouterr()
    assert not err
    assert "file1" in out


def test_no_exclude(capsys, monkeypatch):
    monkeypatch.setattr("sys.argv", ["potodo", "-p", str(REPO_DIR)])
    main()
    out, err = capsys.readouterr()
    assert not err
    assert "file1" in out
    assert "file2" in out
    assert "file3" in out


def test_exclude_file(capsys, monkeypatch):
    monkeypatch.setattr(
        "sys.argv", ["potodo", "-p", str(REPO_DIR), "--exclude", "file*"]
    )
    main()
    out, err = capsys.readouterr()
    assert not err
    assert "file1" not in out
    assert "file2" not in out
    assert "file3" not in out
    assert "excluded" in out  # The only one not being named file


def test_exclude_directory(capsys, monkeypatch):
    monkeypatch.setattr(
        "sys.argv", ["potodo", "-p", str(REPO_DIR), "--exclude", "excluded/*"]
    )
    main()
    out, err = capsys.readouterr()
    assert not err
    assert "file1" in out
    assert "file2" in out
    assert "file3" in out
    assert "file4" not in out  # in the excluded/ directory
    assert "excluded/" not in out


def test_exclude_single_file(capsys, monkeypatch):
    monkeypatch.setattr(
        "sys.argv", ["potodo", "-p", str(REPO_DIR), "--exclude", "file2.po"]
    )
    main()
    out, err = capsys.readouterr()
    assert not err
    assert "file1" in out
    assert "file2" not in out
    assert "file3" in out
    assert "file4" in out



def test_potodo_no_args(run_potodo):
    output = run_potodo([]).out
    assert "# excluded (50.00% done)" in output
    assert "# folder (33.33% done)" in output
    assert "- excluded.po                      1 /   2 ( 50.0% translated)" in output
    assert "- file3.po                         0 /   1 (  0.0% translated)" in output
    assert "# repository (25.00% done)" in output
    assert (
        "- file1.po                         1 /   3 ( 33.0% translated), 1 fuzzy"
        in output
    )


def test_potodo_exclude(run_potodo):
    output = run_potodo(["--exclude", "excluded/", "excluded.po"]).out
    output_short = run_potodo(["-e", "excluded/", "excluded.po"]).out
    assert output == output_short
    assert "# excluded (50.00% done)" not in output
    assert (
        "- excluded.po                      1 /   2 ( 50.0% translated)" not in output
    )
    assert "# repository (25.00% done)" in output
    assert (
        "- file1.po                         1 /   3 ( 33.0% translated), 1 fuzzy"
        in output
    )


def test_potodo_above(run_potodo):
    output = run_potodo(["--above", "40"]).out
    output_short = run_potodo(["-a", "40"]).out
    assert output == output_short
    assert (
        "- file1.po                         1 /   3 ( 33.0% translated), 1 fuzzy"
        not in output
    )
    assert "- excluded.po                      1 /   2 ( 50.0% translated)" in output


def test_potodo_below(run_potodo):
    output = run_potodo(["--below", "40"]).out
    output_short = run_potodo(["-b", "40"]).out
    assert output == output_short
    assert (
        "- file1.po                         1 /   3 ( 33.0% translated), 1 fuzzy"
        in output
    )
    assert (
        "- excluded.po                      1 /   2 ( 50.0% translated)" not in output
    )


def test_potodo_onlyfuzzy(run_potodo):
    output = run_potodo(["--only-fuzzy"]).out
    output_short = run_potodo(["-f"]).out
    assert output == output_short
    assert (
        "- file1.po                         1 /   3 ( 33.0% translated), 1 fuzzy"
        in output
    )
    assert (
        "- excluded.po                      1 /   2 ( 50.0% translated)" not in output
    )


def test_potodo_counts(run_potodo):
    output = run_potodo(["--counts"]).out
    output_short = run_potodo(["-c"]).out
    assert output == output_short
    assert (
        "- excluded.po                      1 /   2 ( 50.0% translated)" not in output
    )
    assert "- file4.po                         1 to do" in output
    assert "# repository (25.00% done)" in output
    assert "- file1.po                         2 to do, 1 fuzzy." in output


def test_potodo_exclude_fuzzy(run_potodo):
    output = run_potodo(["--exclude-fuzzy"]).out
    assert "- excluded.po                      1 /   2 ( 50.0% translated)" in output
    assert "- file1.po                         2 to do, 1 fuzzy." not in output


def test_potodo_matching_files_solo(run_potodo):
    output = run_potodo(["--matching-files"]).out
    output_short = run_potodo(["-l"]).out
    assert output == output_short
    assert "excluded/file4.po" in output
    assert "folder/excluded.po" in output
    assert "folder/file3.po" in output
    assert "file1.po" in output
    assert "file2.po" in output


def test_potodo_matching_files_fuzzy(run_potodo):
    output = run_potodo(["--matching-files", "--only-fuzzy"]).out
    output_short = run_potodo(["-l", "-f"]).out
    assert output == output_short
    assert "file1.po" in output


# TODO: Test hide_reserved, offline options, only_reserved, exclude_reserved, show_reservation_dates
# TODO: Test verbose output levels



import json


def test_txt_output(run_potodo):
    captured = run_potodo(["--exclude", "excluded/", "excluded.po"])

    assert "file1.po" in captured.out
    assert "file2.po" in captured.out
    assert "# folder" in captured.out
    assert "file3.po" in captured.out
    assert "1 fuzzy" in captured.out
    assert "2 fuzzy" not in captured.out
    assert "excluded" not in captured.out


def test_output(run_potodo, repo_dir):
    output = json.loads(
        run_potodo(["--json", "--exclude", "excluded/", "excluded.po"]).out
    )

    expected_folder = {
        "name": "folder/",
        "percent_translated": 0.0,
        "files": [
            {
                "name": "folder/file3",
                "path": f"{repo_dir}/folder/file3.po",
                "entries": 1,
                "fuzzies": 0,
                "translated": 0,
                "percent_translated": 0,
                "reserved_by": None,
                "reservation_date": None,
            },
        ],
    }
    expected_repository = {
        "name": "repository/",
        "percent_translated": 25.0,
        "files": [
            {
                "name": "repository/file1",
                "path": f"{repo_dir}/file1.po",
                "entries": 3,
                "fuzzies": 1,
                "translated": 1,
                "percent_translated": 33,
                "reserved_by": None,
                "reservation_date": None,
            },
            {
                "name": "repository/file2",
                "path": f"{repo_dir}/file2.po",
                "entries": 1,
                "fuzzies": 0,
                "translated": 0,
                "percent_translated": 0,
                "reserved_by": None,
                "reservation_date": None,
            },
        ],
    }

    assert len(output) == 2
    for item in output:
        if item["name"] == "repository/":
            assert item == expected_repository
        if item["name"] == "folder/":
            assert item == expected_folder



import argparse
import logging
import os
import sys
from argparse import Namespace
from pathlib import Path

from potodo import __version__


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        prog="potodo",
        description="List and prettify the po files left to translate.",
    )

    parser.add_argument(
        "-p",
        "--path",
        help="execute Potodo in path",
        metavar="path",
    )

    parser.add_argument(
        "-e",
        "--exclude",
        nargs="+",
        default=[],
        help="gitignore-style patterns to exclude from search.",
        metavar="path",
    )

    parser.add_argument(
        "-a",
        "--above",
        default=0,
        metavar="X",
        type=int,
        help="list all TODOs above given X%% completion",
    )

    parser.add_argument(
        "-b",
        "--below",
        default=100,
        metavar="X",
        type=int,
        help="list all TODOs below given X%% completion",
    )

    parser.add_argument(
        "-f",
        "--only-fuzzy",
        dest="only_fuzzy",
        action="store_true",
        help="print only files marked as fuzzys",
    )

    parser.add_argument(
        "-u",
        "--api-url",
        help=(
            "API URL to retrieve reservation tickets (https://api.github.com/repos/ORGANISATION/REPOSITORY/issues?state=open or https://git.afpy.org/api/v1/repos/ORGANISATION/REPOSITORY/issues?state=open&type=issues)"
        ),
    )

    parser.add_argument(
        "-n",
        "--no-reserved",
        dest="hide_reserved",
        action="store_true",
        help="don't print info about reserved files",
    )

    parser.add_argument(
        "-c",
        "--counts",
        action="store_true",
        help="render list with the count of remaining entries "
        "(translate or review) rather than percentage done",
    )

    parser.add_argument(
        "-j",
        "--json",
        action="store_true",
        dest="json_format",
        help="format output as JSON",
    )

    parser.add_argument(
        "--exclude-fuzzy",
        action="store_true",
        dest="exclude_fuzzy",
        help="select only files without fuzzy entries",
    )

    parser.add_argument(
        "--exclude-reserved",
        action="store_true",
        dest="exclude_reserved",
        help="select only files that aren't reserved",
    )

    parser.add_argument(
        "--only-reserved",
        action="store_true",
        dest="only_reserved",
        help="select only only reserved files",
    )

    parser.add_argument(
        "--show-reservation-dates",
        action="store_true",
        dest="show_reservation_dates",
        help="show issue creation dates",
    )

    parser.add_argument(
        "--no-cache",
        action="store_true",
        dest="no_cache",
        help="Disables cache (Cache is disabled when files are modified)",
    )

    parser.add_argument(
        "-i",
        "--interactive",
        action="store_true",
        dest="is_interactive",
        help="Activates the interactive menu",
    )

    parser.add_argument(
        "-l",
        "--matching-files",
        action="store_true",
        dest="matching_files",
        help="Suppress normal output; instead print the name of each matching po file from which output would normally "
        "have been printed.",
    )

    parser.add_argument(
        "--version", action="version", version="%(prog)s " + __version__
    )

    parser.add_argument(
        "-v", "--verbose", action="count", default=0, help="Increases output verbosity"
    )

    # Initialize args and check consistency
    args = parser.parse_args()
    check_args(args)
    return args


def check_args(args: Namespace) -> None:
    # If below is lower than above, raise an error
    if args.below < args.above:
        print(
            "Potodo: 'below' value must be greater than 'above' value.", file=sys.stderr
        )
        sys.exit(1)

    if args.json_format and args.is_interactive:
        print(
            "Potodo: Json format and interactive modes cannot be activated at the same time.",
            file=sys.stderr,
        )
        sys.exit(1)

    if args.is_interactive:
        try:
            import termios  # noqa
        except ImportError:
            import platform

            print(
                f'Potodo: "{platform.system()}" is not supported for interactive mode',
                file=sys.stderr,
            )
            sys.exit(1)

    if args.exclude_fuzzy and args.only_fuzzy:
        print(
            "Potodo: Cannot pass --exclude-fuzzy and --only-fuzzy at the same time.",
            file=sys.stderr,
        )
        sys.exit(1)

    if args.exclude_reserved and args.only_reserved:
        print(
            "Potodo: Cannot pass --exclude-reserved and --only-reserved at the same time.",
            file=sys.stderr,
        )
        sys.exit(1)

    # If no path is specified, use current directory
    if not args.path:
        args.path = os.getcwd()

    args.path = Path(args.path).resolve()

    try:
        levels = [logging.CRITICAL, logging.WARNING, logging.INFO, logging.DEBUG]
        args.logging_level = levels[args.verbose]
    except IndexError:
        print("Too many `-v`, what do you think you'll get?", file=sys.stderr)
        sys.exit(1)



def test_potodo_help(run_potodo):
    output = run_potodo(["--help"]).out
    output_short = run_potodo(["-h"]).out
    assert output == output_short
    assert "-h, --help            show this help message and exit" in output


def test_potodo_above_below_conflict(run_potodo):
    output = run_potodo(["--above", "50", "--below", "40"]).err
    output_short = run_potodo(["-a", "50", "-b", "40"]).err
    assert (
        output
        == output_short
        == "Potodo: 'below' value must be greater than 'above' value.\n"
    )


def test_potodo_json_interactive_conflict(run_potodo):
    output = run_potodo(["--json", "--interactive"]).err
    output_short = run_potodo(["-j", "-i"]).err
    assert (
        output
        == output_short
        == "Potodo: Json format and interactive modes cannot be activated at the same time.\n"
    )


def test_potodo_exclude_and_only_fuzzy_conflict(run_potodo):
    output = run_potodo(["--exclude-fuzzy", "--only-fuzzy"]).err
    assert (
        output
        == "Potodo: Cannot pass --exclude-fuzzy and --only-fuzzy at the same time.\n"
    )


def test_potodo_exclude_and_only_reserved_conflict(run_potodo):
    output = run_potodo(["--exclude-reserved", "--only-reserved"]).err
    assert (
        output
        == "Potodo: Cannot pass --exclude-reserved and --only-reserved at the same time.\n"
    )



import itertools
import logging
import os
import pickle
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Any, Callable, Dict, List, Optional, Sequence, Set, cast

import polib

from potodo import __version__ as VERSION


class PoFileStats:
    """Statistics about a po file.

    Contains all the necessary information about the progress of a given po file.

    Beware this file is pickled (for the cache), don't store actual
    entries in its __dict__, just stats.
    """

    def __init__(self, path: Path):
        """Initializes the class with all the correct information"""
        self.path: Path = path
        self.filename: str = path.name
        self.mtime = os.path.getmtime(path)
        self.directory: str = self.path.parent.name
        self.reserved_by: Optional[str] = None
        self.reservation_date: Optional[str] = None
        self.filename_dir: str = self.directory + "/" + self.filename
        self.stats: Dict[str, int] = {}

    def __eq__(self, other: object) -> bool:
        return isinstance(other, type(self)) and self.path == other.path

    def __hash__(self) -> int:
        return hash(("PoFileStats", self.path))

    @property
    def fuzzy(self) -> int:
        self.parse()
        return self.stats["fuzzy"]

    @property
    def translated(self) -> int:
        self.parse()
        return self.stats["translated"]

    @property
    def untranslated(self) -> int:
        self.parse()
        return self.stats["untranslated"]

    @property
    def entries(self) -> int:
        self.parse()
        return self.stats["entries"]

    @property
    def percent_translated(self) -> int:
        self.parse()
        return self.stats["percent_translated"]

    def parse(self) -> None:
        if self.stats:
            return  # Stats already computed.
        pofile = polib.pofile(str(self.path))
        self.stats = {
            "fuzzy": len(
                [entry for entry in pofile if entry.fuzzy and not entry.obsolete]
            ),
            "percent_translated": pofile.percent_translated(),
            "entries": len([e for e in pofile if not e.obsolete]),
            "untranslated": len(pofile.untranslated_entries()),
            "translated": len(pofile.translated_entries()),
        }

    def __repr__(self) -> str:
        if self.stats:
            return f"<PoFileStats {self.path!r} {self.entries} entries>"
        return f"<PoFileStats {self.path!r} (unparsed)>"

    def __lt__(self, other: "PoFileStats") -> bool:
        """When two PoFiles are compared, their filenames are compared."""
        return self.path < other.path

    def reservation_str(self, with_reservation_dates: bool = False) -> str:
        if self.reserved_by is None:
            return ""
        as_string = f"reserved by {self.reserved_by}"
        if with_reservation_dates:
            as_string += f" ({self.reservation_date})"
        return as_string

    @property
    def missing(self) -> int:
        return self.fuzzy + self.untranslated

    def as_dict(self) -> Dict[str, Any]:
        return {
            "name": f"{self.directory}/{self.filename.replace('.po', '')}",
            "path": str(self.path),
            "entries": self.entries,
            "fuzzies": self.fuzzy,
            "translated": self.translated,
            "percent_translated": self.percent_translated,
            "reserved_by": self.reserved_by,
            "reservation_date": self.reservation_date,
        }


class PoDirectoryStats:
    """Represent a directory containing multiple `.po` files."""

    def __init__(self, path: Path, files_stats: Sequence[PoFileStats]):
        self.path = path
        self.files_stats = files_stats

    def __repr__(self) -> str:
        return f"<PoDirectoryStats {self.path!r} with {len(self.files_stats)} files>"

    @property
    def translated(self) -> int:
        """Qty of translated entries in the po files of this directory."""
        return sum(po_file.translated for po_file in self.files_stats)

    @property
    def entries(self) -> int:
        """Qty of entries in the po files of this directory."""
        return sum(po_file.entries for po_file in self.files_stats)

    @property
    def completion(self) -> float:
        """Return % of completion of this directory."""
        return 100 * self.translated / self.entries

    def __eq__(self, other: object) -> bool:
        return isinstance(other, type(self)) and self.path == other.path

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, type(self)):
            return False
        return self.path < other.path

    def __le__(self, other: object) -> bool:
        if not isinstance(other, type(self)):
            return False
        return self.path <= other.path

    def __gt__(self, other: object) -> bool:
        if not isinstance(other, type(self)):
            return False
        return self.path > other.path

    def __ge__(self, other: object) -> bool:
        if not isinstance(other, type(self)):
            return False
        return self.path >= other.path


class PoProjectStats:
    """Represents the root of the hierarchy of `.po` files."""

    def __init__(self, path: Path):
        self.path = path
        # self.files can be persisted on disk
        # using `.write_cache()` and `.read_cache()
        self.files: Set[PoFileStats] = set()
        self.excluded_files: Set[PoFileStats] = set()

    def filter(self, filter_func: Callable[[PoFileStats], bool]) -> None:
        """Filter files according to a filter function.

        If filter is applied multiple times, it behave like only last
        filter has been applied.
        """
        all_files = self.files | self.excluded_files
        self.files = set()
        self.excluded_files = set()
        for file in all_files:
            if filter_func(file):
                self.files.add(file)
            else:
                self.excluded_files.add(file)

    @property
    def translated(self) -> int:
        """Qty of translated entries in the po files of this directory."""
        return sum(
            directory_stats.translated for directory_stats in self.stats_by_directory()
        )

    @property
    def entries(self) -> int:
        """Qty of entries in the po files of this directory."""
        return sum(
            directory_stats.entries for directory_stats in self.stats_by_directory()
        )

    @property
    def completion(self) -> float:
        """Return % of completion of this project."""
        return 100 * self.translated / self.entries

    def rescan(self) -> None:
        """Scan disk to search for po files.

        This is the only function that hit the disk.
        """
        for path in list(self.path.rglob("*.po")):
            if PoFileStats(path) not in self.files:
                self.files.add(PoFileStats(path))

    def stats_by_directory(self) -> List[PoDirectoryStats]:
        return [
            PoDirectoryStats(directory, list(po_files))
            for directory, po_files in itertools.groupby(
                sorted(self.files, key=lambda po_file: po_file.path.parent),
                key=lambda po_file: po_file.path.parent,
            )
        ]

    def read_cache(self) -> None:
        """Restore all PoFileStats from disk.

        While reading the cache, outdated entires are **not** loaded.
        """
        cache_path = self.path / ".potodo" / "cache.pickle"

        logging.debug("Trying to load cache from %s", cache_path)
        try:
            with open(cache_path, "rb") as handle:
                data = pickle.load(handle)
        except FileNotFoundError:
            logging.warning("No cache found")
            return
        logging.debug("Found cache")
        if data.get("version") != VERSION:
            logging.info("Found old cache, ignored it.")
            return
        for po_file in cast(List[PoFileStats], data["data"]):
            if os.path.getmtime(po_file.path.resolve()) == po_file.mtime:
                self.files.add(po_file)

    def write_cache(self) -> None:
        """Persists all PoFileStats to disk."""
        cache_path = self.path / ".potodo" / "cache.pickle"
        os.makedirs(cache_path.parent, exist_ok=True)
        data = {"version": VERSION, "data": self.files | self.excluded_files}
        with NamedTemporaryFile(
            mode="wb", delete=False, dir=str(cache_path.parent), prefix=cache_path.name
        ) as tmp:
            pickle.dump(data, tmp)
        os.rename(tmp.name, cache_path)
        logging.debug("Wrote PoProjectStats cache to %s", cache_path)



import json
import logging
from pathlib import Path
from typing import Callable, List

from gitignore_parser import rule_from_pattern

from potodo.arguments_handling import parse_args
from potodo.forge_api import get_issue_reservations
from potodo.json import json_dateconv
from potodo.logging import setup_logging
from potodo.po_file import PoFileStats, PoProjectStats


def scan_path(
    path: Path,
    no_cache: bool,
    hide_reserved: bool,
    api_url: str,
) -> PoProjectStats:
    logging.debug("Finding po files in %s", path)
    po_project = PoProjectStats(path)

    if no_cache:
        logging.debug("Creating PoFileStats objects for each file without cache")
    else:
        po_project.read_cache()

    po_project.rescan()

    if api_url and not hide_reserved:
        issue_reservations = get_issue_reservations(api_url)
        for po_file_stats in po_project.files:
            reserved_by, reservation_date = issue_reservations.get(
                po_file_stats.filename_dir.lower(), (None, None)
            )
            if reserved_by and reservation_date:
                po_file_stats.reserved_by = reserved_by
                po_file_stats.reservation_date = reservation_date
            else:  # Just in case we remember it's reserved from the cache:
                po_file_stats.reserved_by = None
                po_file_stats.reservation_date = None

    return po_project


def print_matching_files(po_project: PoProjectStats) -> None:
    for directory_stats in sorted(po_project.stats_by_directory()):
        for file_stat in sorted(directory_stats.files_stats):
            print(file_stat.path)


def print_po_project(
    po_project: PoProjectStats, counts: bool, show_reservation_dates: bool
) -> None:
    for directory_stats in sorted(po_project.stats_by_directory()):
        print(
            f"\n\n# {directory_stats.path.name} ({directory_stats.completion:.2f}% done)\n"
        )

        for file_stat in sorted(directory_stats.files_stats):
            line = f"- {file_stat.filename:<30} "
            if counts:
                line += f"{file_stat.missing:3d} to do"
            else:
                line += f"{file_stat.translated:3d} / {file_stat.entries:3d}"
                line += f" ({file_stat.percent_translated:5.1f}% translated)"
            if file_stat.fuzzy:
                line += f", {file_stat.fuzzy} fuzzy"
            if file_stat.reserved_by is not None:
                line += ", " + file_stat.reservation_str(show_reservation_dates)
            print(line + ".")

    if po_project.entries != 0:
        print(f"\n\n# TOTAL ({po_project.completion:.2f}% done)\n")


def print_po_project_as_json(po_project: PoProjectStats) -> None:
    print(
        json.dumps(
            [
                {
                    "name": f"{directory_stats.path.name}/",
                    "percent_translated": directory_stats.completion,
                    "files": [
                        po_file.as_dict()
                        for po_file in sorted(directory_stats.files_stats)
                    ],
                }
                for directory_stats in sorted(po_project.stats_by_directory())
            ],
            indent=4,
            separators=(",", ": "),
            sort_keys=False,
            default=json_dateconv,
        )
    )


def build_ignore_matcher(path: Path, exclude: List[str]) -> Callable[[str], bool]:
    path = path.resolve()
    potodo_ignore = path / ".potodoignore"
    rules = []
    if potodo_ignore.exists():
        for line in potodo_ignore.read_text().splitlines():
            rule = rule_from_pattern(line, path)
            if rule:
                rules.append(rule)
    rules.append(rule_from_pattern(".git/", path))
    for rule in exclude:
        rules.append(rule_from_pattern(rule, path))
    return lambda file_path: any(r.match(file_path) for r in rules)


def main() -> None:
    args = parse_args()

    if args.logging_level:
        setup_logging(args.logging_level)

    logging.info("Logging activated.")
    logging.debug("Executing potodo with args %s", args)

    ignore_matches = build_ignore_matcher(args.path, args.exclude)

    def select(po_file: PoFileStats) -> bool:
        """Return True if the po_file should be displayed, False otherwise."""
        if ignore_matches(str(po_file.path)):
            return False
        if args.only_fuzzy and not po_file.fuzzy:
            return False
        if args.exclude_fuzzy and po_file.fuzzy:
            return False
        if (
            po_file.percent_translated == 100
            or po_file.percent_translated < args.above
            or po_file.percent_translated > args.below
        ):
            return False

        # unless the offline/hide_reservation are enabled
        if args.exclude_reserved and po_file.reserved_by:
            return False
        if args.only_reserved and not po_file.reserved_by:
            return False

        return True

    if args.is_interactive:
        from potodo.interactive import interactive_output

        interactive_output(args.path, ignore_matches)
        return

    po_project = scan_path(args.path, args.no_cache, args.hide_reserved, args.api_url)
    po_project.filter(select)
    if args.matching_files:
        print_matching_files(po_project)
    elif args.json_format:
        print_po_project_as_json(po_project)
    else:
        print_po_project(po_project, args.counts, args.show_reservation_dates)
    po_project.write_cache()
